/*
								+---------------------------------+
								|                                 |
								|   ***   Advanced effect   ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|           Effect.cpp            |
								|                                 |
								+---------------------------------+
*/

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "../UberLame_src/OpenGL20.h"
#include <gl/glu.h>
#include "../UberLame_src/OpenGLDrv.h"
#include "../UberLame_src/OpenGLState.h"
#include "../UberLame_src/Timer.h"
#include "../UberLame_src/Vector.h"
//#include "../UberLame_src/StrokeFont.h"
#include "../UberLame_src/Texture.h"
#include "../UberLame_src/RenderBuffer2.h"
#include "../UberLame_src/Shader2.h"
#include "../UberLame_src/ShaderSys.h"
#include "../UberLame_src/Jpeg.h"
#include "../UberLame_src/Mersene.h"
#include "resource.h"
#include "Dialogs.h"
#include "Effect.h"

static CGLTexture_2D *p_Load_JpegTexture(const char *p_s_filename, CGLState *p_state, bool b_mipmaps)
{
	HRSRC h_res;
	if(!(h_res = FindResource(NULL, p_s_filename, "JPG")))
		return 0;
	FILE *p_fw;
	if(!(p_fw = fopen("_tmp.jpg", "wb")))
		return 0;
	int n_size = SizeofResource(NULL, h_res);
	void *p_data = LockResource(LoadResource(NULL, h_res));
	if(fwrite(p_data, 1, n_size, p_fw) != n_size) {
		fclose(p_fw);
		return 0;
	}
	fclose(p_fw);
	p_s_filename = "_tmp.jpg";
	// hack - write resource into temporary file

	FILE *p_fr;
	if(!(p_fr = fopen(p_s_filename, "rb"))) {
		DeleteFile("_tmp.jpg");
		return 0;
	}

	TBmp *p_bitmap;
	CTinyJpegDecoder jpeg_dec;
	if(!(p_bitmap = jpeg_dec.p_Decode_JPEG(p_fr))) {
		fclose(p_fr);
		DeleteFile("_tmp.jpg");
		return 0;
	}
	fclose(p_fr);
	DeleteFile("_tmp.jpg");

	CGLTexture_2D *p_texture;
	if(!(p_texture = new(std::nothrow) CGLTexture_2D(p_state, p_bitmap->n_width, p_bitmap->n_height, GL_RGBA8,
	   b_mipmaps, 0, /*GL_TEXTURE_2D,*/ GL_BGRA, GL_UNSIGNED_BYTE, p_bitmap->p_buffer))) {
		delete[] p_bitmap->p_buffer;
		delete p_bitmap;
		return 0;
	}
	delete[] p_bitmap->p_buffer;
	delete p_bitmap;

	//p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	//p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);

	if(!p_texture->b_Status() || glGetError() != GL_NO_ERROR) {
		delete p_texture;
		return 0;
	}

	return p_texture;
}

/*
 *								=== CSmokeBackground ===
 */

/*
 *	CSmokeBackground::CSmokeBackground()
 *		- default constructor
 */
CSmokeBackground::CSmokeBackground()
	:m_p_smoke_texture(0)
{
}

/*
 *	virtual CSmokeBackground::~CSmokeBackground()
 *		- default destructor (supposed to call Shutdown())
 */
CSmokeBackground::~CSmokeBackground()
{
	if(b_Status()) {
		CGLState cur_state/*(wglGetCurrentDC(), wglGetCurrentContext())*/;
		//cur_state.Update_States();
		Shutdown(&cur_state);
	}
}

/*
 *	virtual const char *CSmokeBackground::p_s_Name()
 *		- returns short effect name (c-like identifier, name of effect class for example)
 */
const char *CSmokeBackground::p_s_Name()
{
	return "CSmokeBackground";
}

/*
 *	virtual const char *CSmokeBackground::p_s_LongName()
 *		- returns full effect name
 */
const char *CSmokeBackground::p_s_LongName()
{
	return "smoke background";
}

/*
 *	virtual bool CSmokeBackground::b_Supported()
 *		- returns true in case effect is supported by OpenGL, otherwise false
 */
bool CSmokeBackground::b_Supported()
{
	return true;
}

/*
 *	virtual bool CSmokeBackground::Init(CGLState *p_state, int n_width, int n_height)
 *		- loads all effect resources, creates OpenGL buffers, compiles shaders, etc
 *		- returns true on success, false on failure
 */
bool CSmokeBackground::Init(CGLState *p_state, /*CLoggerDialog &r_logwnd,*/ int n_width, int n_height)
{
	//r_logwnd.AddLine("initializing smoke effect ...");
	//r_logwnd.AddLine("\tloading texture from resource section ...");
	if(!(m_p_smoke_texture = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG1)/*"f:\\windows\\cloud.jpg"*/, p_state, false)))
		return false;
	//r_logwnd.AddLine("\t\tok"); 
	return true;
}

/*
 *	virtual bool CSmokeBackground::b_Status()
 *		- returns true in case effect is loaded and ready for rendering, otherwise false
 */
bool CSmokeBackground::b_Status()
{
	return m_p_smoke_texture != 0;
}

/*
 *	virtual bool CSmokeBackground::Shutdown(CGLState *p_state)
 *		- shutdown effect
 *		- returns true in case no errors occured (even in case effect was not initialized)
 *		  otherwise returns false
 */
bool CSmokeBackground::Shutdown(CGLState *p_state)
{
	if(m_p_smoke_texture) {
		delete m_p_smoke_texture;
		m_p_smoke_texture = 0;
	}
	return true;
}

/*
 *	virtual float CSmokeBackground::f_Length()
 *		- returns effect length in seconds
 */
float CSmokeBackground::f_Length()
{
	return 50;
}

/*
 *	virtual bool CSmokeBackground::Render(CGLState *p_state, float f_time)
 *		- renders a single frame into current framebuffer; f_time is frame time,
 *		  it can be out of <0, f_Length()> range, it's up to effect on how to handle that
 *		  (repeat itself, fade out, render nothing ...)
 */
bool CSmokeBackground::Render(CGLState *p_state, float f_time)
{
	const float f_particle_size = .75f;
	const int n_particle_num = 128;

	if(!m_p_smoke_texture)
		return false;

	static float p_speed_table[4 * n_particle_num];
	static bool b_first_time = true;

	if(b_first_time) {
		CMerseneTwister tw;
		tw.init_genrand(0);
		float *p_speed = p_speed_table;
		for(int i = 0; i < n_particle_num; ++ i, p_speed += 4) {
			p_speed[0] = 1 - tw.genrand_real1() * 2;
			p_speed[1] = .1f + .3f * tw.genrand_real1();
			p_speed[2] = tw.genrand_real1() * 4;
			p_speed[3] = 1 - tw.genrand_real1() * 2;
		}
	}
	// init speed table (twister is costy, want to have minimal workload per frame)

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	m_p_smoke_texture->Bind_Enable(p_state);
	p_state->EnableBlend();
	p_state->BlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
	p_state->DisableDepthTest();
	p_state->DisableCullFace();
	glColor3f(.0625f, .0625f, .0625f);

	glBegin(GL_QUADS);
	float *p_speed = p_speed_table;
	for(int i = 0; i < n_particle_num; ++ i, p_speed += 4) {
		float f_x = -1 + i / (float)(n_particle_num / 2) + p_speed[0];
		float f_y = 1.5f - fmod(p_speed[2] + f_time * p_speed[1], 3);
		float f_t = p_speed[3] * f_time;

		glTexCoord2f(0, 0);
		glVertex2f(f_x + f_particle_size * (float)sin(f_t),
			f_y + f_particle_size * (float)cos(f_t));
		glTexCoord2f(0, 1);
		glVertex2f(f_x + f_particle_size * (float)sin(f_t + f_pi * .5f),
			f_y + f_particle_size * (float)cos(f_t + f_pi * .5f));
		glTexCoord2f(1, 1);
		glVertex2f(f_x + f_particle_size * (float)sin(f_t + f_pi),
			f_y + f_particle_size * (float)cos(f_t + f_pi));
		glTexCoord2f(1, 0);
		glVertex2f(f_x + f_particle_size * (float)sin(f_t + f_pi * 1.5f),
			f_y + f_particle_size * (float)cos(f_t + f_pi * 1.5f));
	}
	glEnd();
	// draw screen-aligned, rotating quads

	return true;
}

/*
 *								=== ~CSmokeBackground ===
 */

/*
 *								=== CHexLattice ===
 */

const int CHexLattice::m_p_rod_index_array_top[] = {0, 1, 2, 3, 4, 5};
const int CHexLattice::m_p_rod_index_array_side[] = {6 + 0, 6 + 12, 6 + 13, 6 + 1,
													 6 + 2, 6 + 14, 6 + 15, 6 + 3,
													 6 + 4, 6 + 16, 6 + 17, 6 + 5,
													 6 + 6, 6 + 18, 6 + 19, 6 + 7,
													 6 + 8, 6 + 20, 6 + 21, 6 + 9,
													 6 + 10, 6 + 22, 6 + 23, 6 + 11
};
const float CHexLattice::f_rod_radius = 1.5f;
const float CHexLattice::f_rod_height = 5.0f;

/*TShaderInfo::TShader CHexLattice::m_t_vs_high, CHexLattice::m_t_fs_high, CHexLattice::m_t_fs_high2;
TShaderInfo::TShader::TSamplerInfo CHexLattice::m_p_fs_samplers[3], CHexLattice::m_p_fs_samplers2[1];*/

TShaderInfo CHexLattice::m_t_vertex_shader = TShaderInfo("local_bump_vs", false,
	"varying vec3 v_to_light, v_to_eye;\n"
	"varying float f_inv_dist;\n"
	"\n"
	"void main()\n"
	"{\n"
	"    vec3 v_worldspace_pos = (gl_ModelViewMatrix * gl_Vertex).xyz;\n"
	"    gl_Position = ftransform();\n"
	"    gl_TexCoord[0] = gl_MultiTexCoord0;\n"
	"\n"
	"    vec3 v_to_light_ws = gl_LightSource[0].position.xyz - v_worldspace_pos;\n"
	"    vec3 v_to_eye_ws = -v_worldspace_pos;\n"
	"    f_inv_dist = gl_DepthRange.diff / (length(v_to_light_ws) - gl_DepthRange.near);"
	"\n"
	"    vec3 v_normal = normalize(gl_NormalMatrix * gl_Normal);\n"
	"    vec3 v_tangent = normalize(gl_NormalMatrix * gl_Color.xyz);\n"
	"    vec3 v_bitangent = cross(v_tangent, v_normal);\n"
	"\n"
	"    v_to_light.x = dot(v_to_light_ws, v_tangent);\n"
	"    v_to_light.y = dot(v_to_light_ws, v_bitangent);\n"
	"    v_to_light.z = dot(v_to_light_ws, v_normal);\n"
	"\n"
	"    v_to_eye.x = dot(v_to_eye_ws, v_tangent);\n"
	"    v_to_eye.y = dot(v_to_eye_ws, v_bitangent);\n"
	"    v_to_eye.z = dot(v_to_eye_ws, v_normal);\n"
	"\n"
	"    gl_FogFragCoord = gl_DepthRange.diff / (max(gl_DepthRange.near - v_worldspace_pos.z + 7.0, 0.0001));\n;"
	"}\n");

TShaderInfo CHexLattice::m_t_fragment_shader = TShaderInfo("local_bump_fs", true,
	"uniform sampler2D local;\n"
	"uniform sampler2D specular;\n"
	"uniform sampler2D diffuse;\n"
	"\n"
	"varying vec3 v_to_light, v_to_eye;\n"
	"varying float f_inv_dist;\n"
	"\n"
	"void main()\n"
	"{\n"
	"    const vec3 v_half = vec3(.5, .5, .5);\n"
	"\n"
	"    float f_atten = clamp(1.0 - gl_LightSource[0].constantAttenuation -\n"
	"        gl_LightSource[0].linearAttenuation / f_inv_dist, 0.0, 1.0);\n"
	"    if(f_atten > .0039) {\n" // 1 / 256
	"        vec3 v_real_normal = normalize(texture2D(local, gl_TexCoord[0].xy).xyz - v_half)\n;"
	"        vec3 v_light = normalize(v_to_light.xyz).xyz;\n"
	"        vec3 v_eye = normalize(v_to_eye.xyz).xyz;\n;"
	"\n"
	"        float f_diffuse = clamp(dot(v_light, v_real_normal), 0.0, 1.0);\n"
	"        float f_specular = clamp(dot(reflect(v_light, v_real_normal), v_eye), 0.0, 1.0);\n"
	"        f_specular *= f_specular;\n"
	"        f_specular *= f_specular;\n"
	"        f_specular *= f_specular;\n"
	"\n"
	"        f_specular *= f_atten;\n"
	"        f_diffuse *= f_atten;\n"
	"\n"
	"        gl_FragColor.xyz = (gl_LightSource[0].ambient.xyz +\n"
	"            f_diffuse * (gl_LightSource[0].diffuse * texture2D(diffuse, gl_TexCoord[0].xy)).xyz +\n"
	"            f_specular * (gl_LightSource[0].specular * texture2D(specular, gl_TexCoord[0].xy)).xyz);\n"
	"    } else\n"
	"        gl_FragColor.xyz = vec3(.0, .0, .0);\n"
	"    gl_FragColor.w = 1.0 - clamp(.1 / gl_FogFragCoord, .0, 1.0);\n"
	"}\n", "diffuse|specular|local");

TShaderInfo CHexLattice::m_t_fragment_shader2 = TShaderInfo("depth_of_field_postprocessing", true,
	"uniform sampler2D texture;\n"
	"uniform vec2 v_direction;\n"
	"\n"
	"void main()\n"
	"{\n"
	"    float f_sample = texture2D(texture, gl_TexCoord[0].xy).w;\n"
	"\n"
	"    float f_sample_num = min(1.0 + ((1.0 - f_sample) * 15.0), 1.0 + f_sample * 15.0);\n" // 1 - 16 samples
	"    vec4 v_accum = vec4(.0, .0, .0, .0);\n"
	"    vec2 v_dir = v_direction * (1.0 - (1.0 - fract(f_sample_num)) / f_sample_num);\n"
	"    float f_bias = (1.0 - f_sample) * 4.0;\n" // have 3 mip-maps
	"    v_dir *= f_bias;\n"
	"    vec2 v_texcoord = gl_TexCoord[0].xy - v_dir * (f_sample_num * .5);"
	"    float f = f_sample_num;\n"
	"    float a = .0;"
	"    do {\n"
	"        v_accum += texture2D(texture, v_texcoord, f_bias);\n"
	"        v_texcoord += v_dir;\n"
	"        a += 1.0;"
	"    } while(-- f > -.5);\n"
	"    gl_FragColor = v_accum / a;\n"
	"    if(v_direction.y == .0)\n" // for 2nd pass only
	"        gl_FragColor.w = f_sample;\n" // needed for 2nd pass
	"}\n");

const bool CHexLattice::m_b_enable_mipmaps = true;
const float f_downscale_render_buffer = 1.0f;

/*
 *	CHexLattice::CHexLattice()
 *		- default constructor
 */
CHexLattice::CHexLattice()
	:m_p_rod_top_diffuse(0), m_p_rod_top_specular(0), m_p_rod_top_local(0),
	m_p_rod_side_diffuse(0), m_p_rod_side_specular(0), m_p_rod_side_local(0),
	m_b_dl_ready(false), m_p_postprocessing_shader(0), m_p_local_bump_shader(0),
	m_p_render_buffer(0), m_t_direction_param(0, 0, 0, 0)
{
	m_p_render_texture[0] = 0;
	m_p_render_texture[1] = 0;
}

/*
 *	virtual CHexLattice::~CHexLattice()
 *		- default destructor (supposed to call Shutdown())
 */
CHexLattice::~CHexLattice()
{
	if(b_Status()) {
		CGLState cur_state/*(wglGetCurrentDC(), wglGetCurrentContext())*/;
		//cur_state.Update_States();
		Shutdown(&cur_state);
	}
	m_b_dl_ready = false;
}

/*
 *	virtual const char *CHexLattice::p_s_Name()
 *		- returns short effect name (c-like identifier, name of effect class for example)
 */
const char *CHexLattice::p_s_Name()
{
	return "CHexLattice";
}

/*
 *	virtual const char *CHexLattice::p_s_LongName()
 *		- returns full effect name
 */
const char *CHexLattice::p_s_LongName()
{
	return "hexagonal rod lattice";
}

/*
 *	virtual bool CHexLattice::b_Supported()
 *		- returns true in case effect is supported by OpenGL, otherwise false
 */
bool CHexLattice::b_Supported()
{
	return true;
}

/*
 *	virtual bool CHexLattice::Init(CGLState *p_state, int n_width, int n_height)
 *		- loads all effect resources, creates OpenGL buffers, compiles shaders, etc
 *		- returns true on success, false on failure
 */
bool CHexLattice::Init(CGLState *p_state, /*CLoggerDialog &r_logwnd,*/ int n_width, int n_height)
{
	//r_logwnd.AddLine("initializing lattice effect ...");

	m_f_aspect = (float)n_width / n_height;
	m_b_dl_ready = false;
	Generate_RodVertices();
	// some cpu stuff

	//r_logwnd.AddLine("\tgeometry created");

	m_n_width = n_width;
	m_n_height = n_height;

	n_width = (int)(n_width * f_downscale_render_buffer);
	n_height = (int)(n_height * f_downscale_render_buffer);

	if(!CGLExtensionHandler::b_SupportedExtension("GL_ARB_texture_non_power_of_two")) {
		int n_max_tex_size;
		glGetIntegerv(GL_MAX_TEXTURE_SIZE, &n_max_tex_size);
		n_width = (int)ceil(log(n_width) / log(2));
		n_width = (n_width <= 0)? 1 : (((1 << (n_width)) >
			n_max_tex_size)? n_max_tex_size : 1 << (n_width));
		n_height = (int)ceil(log(n_height) / log(2));
		n_height = (n_height <= 0)? 1 : (((1 << (n_height)) >
			n_max_tex_size)? n_max_tex_size : 1 << (n_height));
		//r_logwnd.AddLineF("\tNPOT's not supported, forcing %d  %d (beware aliasing)", n_width, n_height);
	}

	//r_logwnd.AddLine("\tcreating render-buffer (RGBA8, 24bpp depth, 0bpp stencil)");

	if(!(m_p_render_buffer = new(std::nothrow) CGLFrameBuffer_FBO(n_width, n_height, 1, true, GL_RGBA, 8,
	   false, true, false, 24, false, false, 0)) || !m_p_render_buffer->b_Status())
		return false;

	//r_logwnd.AddLine("\t\tok");
	//r_logwnd.AddLine("\tcreating ping-pong textures");

	for(int i = 0; i < 2; ++ i) {
		if(!(m_p_render_texture[i] = new(std::nothrow) CGLTexture_2D(p_state, n_width, n_height, GL_RGBA8)) ||
		   !m_p_render_texture[i]->b_Status())
			return false;
		m_p_render_texture[i]->Bind_Enable(p_state);
		m_p_render_texture[i]->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
		m_p_render_texture[i]->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
		if(m_b_enable_mipmaps)
			m_p_render_texture[i]->r_Parameters(p_state).Set_Texture_Max_Level(3);
		else
			m_p_render_texture[i]->r_Parameters(p_state).Set_Texture_Min_Filter(GL_LINEAR);
	}
	// create renderbuffer and texture to be rendered to

	//r_logwnd.AddLine("\t\tok");
	//r_logwnd.AddLine("\tloading textures from resource section");

	if(!(m_p_rod_top_diffuse = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG2)/*"f:\\windows\\rod_top_d.jpg"*/, p_state, true)) ||
	   !(m_p_rod_top_specular = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG3)/*"f:\\windows\\rod_top_s.jpg"*/, p_state, true)) ||
	   !(m_p_rod_top_local = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG4)/*"f:\\windows\\rod_top_l.jpg"*/, p_state, true)) ||
	   !(m_p_rod_side_diffuse = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG5)/*"f:\\windows\\rod_side_d.jpg"*/, p_state, true)) ||
	   !(m_p_rod_side_specular = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG6)/*"f:\\windows\\rod_side_s.jpg"*/, p_state, true)) ||
	   !(m_p_rod_side_local = p_Load_JpegTexture(MAKEINTRESOURCE(IDR_JPG7)/*"f:\\windows\\rod_side_l.jpg"*/, p_state, true)))
		return false;
	// alloc textures to be used for rendering inside render-buffer
	
	//r_logwnd.AddLine("\t\tok");
	//r_logwnd.AddLine("\tcompilling shaders");

	for(int i = 0; i < 2; ++ i) {
		CGLShaderObject *&r_p_shader = (i)? m_p_postprocessing_shader : m_p_local_bump_shader;
		if(!(r_p_shader = new(std::nothrow) CGLShaderObject((i)? 0 : &m_t_vertex_shader,
		   (i)? &m_t_fragment_shader2 : &m_t_fragment_shader)))
			return false;
		if(!r_p_shader->b_ShaderSupport(true))
			return false;
		char *p_s_compile_log, *p_s_link_log;
		if(!r_p_shader->Compile_Link(p_state, p_s_compile_log, p_s_link_log)) {
			if(p_s_compile_log) {
				MessageBox(0, p_s_compile_log, "CGLShaderObject::Compile", MB_OK);
				delete[] p_s_compile_log;
			}
			if(p_s_link_log) {
				MessageBox(0, p_s_link_log, "CGLShaderObject::Link", MB_OK);
				delete[] p_s_link_log;
			}
			return false;
		}
		if(p_s_compile_log) {
			//MessageBox(0, p_s_compile_log, "CGLShaderObject::Compile", MB_OK);
			delete[] p_s_compile_log;
		}
		if(p_s_link_log) {
			//MessageBox(0, p_s_link_log, "CGLShaderObject::Link", MB_OK);
			delete[] p_s_link_log;
		}
		r_p_shader->Release(p_state);
	}

	//r_logwnd.AddLine("\t\tok");
	m_t_direction_param = m_p_postprocessing_shader->t_Find_Parameter("v_direction");

	return true;
}

/*
 *	virtual bool CHexLattice::b_Status()
 *		- returns true in case effect is loaded and ready for rendering, otherwise false
 */
bool CHexLattice::b_Status()
{
	return m_p_rod_top_diffuse != 0 && m_p_rod_top_specular != 0 && m_p_rod_top_local != 0 &&
		m_p_rod_side_diffuse != 0 && m_p_rod_side_specular != 0 && m_p_rod_side_local != 0 &&
		m_p_local_bump_shader && m_p_local_bump_shader->b_Status() &&
		m_p_postprocessing_shader && m_p_postprocessing_shader->b_Status() &&
		m_p_render_texture[0] && m_p_render_texture[1] && m_p_render_buffer;
}

/*
 *	virtual bool CHexLattice::Shutdown(CGLState *p_state)
 *		- shutdown effect
 *		- returns true in case no errors occured (even in case effect was not initialized)
 *		  otherwise returns false
 */
bool CHexLattice::Shutdown(CGLState *p_state)
{
	if(m_p_rod_top_diffuse) {
		delete m_p_rod_top_diffuse;
		m_p_rod_top_diffuse = 0;
	}
	if(m_p_rod_top_specular) {
		delete m_p_rod_top_specular;
		m_p_rod_top_specular = 0;
	}
	if(m_p_rod_top_local) {
		delete m_p_rod_top_local;
		m_p_rod_top_local = 0;
	}
	if(m_p_rod_side_diffuse) {
		delete m_p_rod_side_diffuse;
		m_p_rod_side_diffuse = 0;
	}
	if(m_p_rod_side_specular) {
		delete m_p_rod_side_specular;
		m_p_rod_side_specular = 0;
	}
	if(m_p_rod_side_local) {
		delete m_p_rod_side_local;
		m_p_rod_side_local = 0;
	}
	if(m_p_local_bump_shader) {
		m_p_local_bump_shader->Delete_Shaders();
		delete m_p_local_bump_shader;
		m_p_local_bump_shader = 0;
	}
	if(m_p_postprocessing_shader) {
		m_p_postprocessing_shader->Delete_Shaders();
		delete m_p_postprocessing_shader;
		m_p_postprocessing_shader = 0;
	}
	if(m_p_render_buffer) {
		delete m_p_render_buffer;
		m_p_render_buffer = 0;
	}
	for(int i = 0; i < 2; ++ i) {
		if(m_p_render_texture[i]) {
			delete m_p_render_texture[i];
			m_p_render_texture[i] = 0;
		}
	}
	return true;
}

/*
 *	virtual float CHexLattice::f_Length()
 *		- returns effect length in seconds
 */
float CHexLattice::f_Length()
{
	return 50;
}

/*
 *	virtual bool CHexLattice::Render(CGLState *p_state, float f_time)
 *		- renders a single frame into current framebuffer; f_time is frame time,
 *		  it can be out of <0, f_Length()> range, it's up to effect on how to handle that
 *		  (repeat itself, fade out, render nothing ...)
 */
bool CHexLattice::Render(CGLState *p_state, float f_time)
{
	glViewport(0, 0, m_p_render_buffer->n_Width(), m_p_render_buffer->n_Height());

	if(!m_p_render_buffer->Bind() ||
	   !m_p_render_buffer->BindColorTexture_2D(*m_p_render_texture[0], GL_TEXTURE_2D))
		return false;

	p_state->ClearColor4f(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	Render_Texture(p_state, f_time); // FBO's have shared rendering state

	if(!m_p_render_buffer->CopyTextureData(true) ||
	   !m_p_render_buffer->BindColorTexture_2D(*m_p_render_texture[1], GL_TEXTURE_2D))
		return false;

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	p_state->DisableCullFace();
	p_state->DisableDepthTest();
	p_state->DisableLighting();
	glColor3f(1, 1, 1);

	for(int i = 0;; ++ i) {
		if(!i)
			glClear(GL_COLOR_BUFFER_BIT);
		if(i) {
			p_state->EnableBlend();
			p_state->BlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
		}
		m_p_render_texture[i]->Bind_Enable(p_state);
		if(m_b_enable_mipmaps)
			glGenerateMipmapEXT(GL_TEXTURE_2D);

		/*float p_direction[2] = {
			(i)? 1.0f / m_p_render_buffer->n_Width() : 0,
			(i)? 0 : 1.0f / m_p_render_buffer->n_Height()
		};
		m_p_direction_param->SetValue(p_direction, CSwizzleGenerator::swz_XY);*/
		// works just the same well

		float p_direction[2] = {1.0f / ((i)? m_p_render_buffer->n_Width() : m_p_render_buffer->n_Height()), 0};
		m_t_direction_param.SetValue(p_direction[i], p_direction[(i + 1) % 2]);
		// set shader parameters (blur direction)

		if(!m_p_postprocessing_shader->Bind(p_state))
			return false;

		glBegin(GL_QUADS);
		glTexCoord2f(1, 1);
		glVertex2f(1, 1);
		glTexCoord2f(1, 0);
		glVertex2f(1, -1);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 1);
		glVertex2f(-1, 1);
		glEnd();
		// fullscreen quad (copy texture to framebuffer)

		m_p_postprocessing_shader->Release(p_state);

		if(i)
			break;

		if(!m_p_render_buffer->CopyTextureData(true) || !m_p_render_buffer->Release())
			return false;

		glViewport(0, 0, m_n_width, m_n_height);
	}
	// apply two-pass depth-sensitive blur (fake depth-of-field)

	p_state->DisableAlphaTest();

	return true;
}

bool CHexLattice::Render_Texture(CGLState *p_state, float f_time)
{
	const float f_z_near = .3f;
	const float f_z_far = 1000.0f;
	const float f_fov = 90.0f;
	const float f_aspect = m_f_aspect;

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	float f_w, f_h;
	f_h = tanf(f_fov * .5f / 180 * f_pi) * f_z_near * .5f;
	f_w = f_h * f_aspect;
	glFrustum(-f_w, f_w, -f_h, f_h, f_z_near, f_z_far);
	// set frustum

	gluLookAt(sin(f_time * .05f + f_pi * sin(f_time * .07f)), 10, 12, 0, 4.5f, 0, 0, 1, 0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set modelview matrix to zero (draw rod in the centre)

	p_state->EnableCullFace();
	p_state->EnableDepthTest();
	glColor3f(1, 1, 1);
	p_state->DisableBlend();

	f_time *= .1f;
	Vector3f v_light_pos((float)sin(f_time) * 5, f_rod_height + 2, (float)cos(f_time) * 5);
	Vector3f v_diffuse(1, 1, 1);
	Vector3f v_specular(3, 3, 3);
	Vector3f v_ambient(.0f, .0f, .0f);
	p_state->LightDiffuse(0, &v_diffuse.x);
	p_state->LightSpecular(0, &v_specular.x);
	p_state->LightAmbient(0, &v_ambient.x);
	p_state->LightPosition(0, &v_light_pos.x);
	p_state->LightConstantAttenuation(0, .125f);
	p_state->LightLinearAttenuation(0, .1f);
	p_state->EnableLighting();
	p_state->EnableLight(0);
	// setup light

	if(!m_p_local_bump_shader->Bind(p_state))
		return false;

	for(int i = 0; i < 8 * 24; ++ i) {
		int n_x = i / 8 - 12;
		int n_y = i % 8 - 5;
		int n_x2 = n_x + 5;
		int n_y2 = n_y - 3;
		Draw_Rod(p_state, n_x, n_y, (float)sin(f_time * 2 + sqrt(n_x * n_x + n_y * n_y)) +
			(float)sin(f_time * 1.5f + sqrt(n_x2 * n_x2 + n_y2 * n_y2)));
	}
	// draw shifting rods

	m_p_local_bump_shader->Release(p_state);

	p_state->DisableLighting();
	p_state->DisableTexture2D();
	glColor3f(0, 0, 0);

	glBegin(GL_QUADS);
	glVertex3f(50, 0, 50);
	glVertex3f(50, 0, 0);
	glVertex3f(-50, 0, 0);
	glVertex3f(-50, 0, 50);
	glEnd();
	// black ground plane

	return true;
}

void CHexLattice::Generate_RodVertices()
{
	TVertex p_vertex[30];

	{int n_index = 0;
	for(TVertex *p_vert_ptr = p_vertex, *p_end = p_vertex + 6;
	   p_vert_ptr < p_end; ++ p_vert_ptr, ++ n_index) {
		float f_angle = n_index * f_pi * .33333f;
		p_vert_ptr->v_position.x = f_rod_radius * (float)sin(f_angle);
		p_vert_ptr->v_position.y = f_rod_height;
		p_vert_ptr->v_position.z = f_rod_radius * (float)cos(f_angle);
		// calc positions

		p_vert_ptr->v_normal = Vector3f(0, 1, 0);
		p_vert_ptr->v_tangent = Vector3f(1, 0, 0);
		p_vert_ptr->v_bitangent = Vector3f(0, 0, 1);
		// normal vectors

		p_vert_ptr->v_texcoord = Vector2f(.5f + .5f * (float)sin(f_angle),
			.5f + .5f * (float)cos(f_angle));
		// texcoords
	}}
	// cap

	{int n_index = 0;
	for(TVertex *p_vert_ptr = p_vertex + 6, *p_end = p_vertex + 30;
	   p_vert_ptr < p_end; ++ p_vert_ptr, ++ n_index) {
		int n_side_index = (n_index % 12) / 2;
		float f_side_angle = n_side_index * f_pi * .33333f + f_pi * .16667;
		// for normal calculation

		int n_vertex_index = (((n_index + 1) / 2) % 6);
		float f_angle = n_vertex_index * f_pi * .33333f;
		// for position calculation

		p_vert_ptr->v_position.x = f_rod_radius * (float)sin(f_angle);
		p_vert_ptr->v_position.y = (n_index < 12)? f_rod_height : 0;
		p_vert_ptr->v_position.z = f_rod_radius * (float)cos(f_angle);
		// calc positions

		p_vert_ptr->v_normal = Vector3f((float)sin(f_side_angle), 0, (float)cos(f_side_angle));
		p_vert_ptr->v_tangent = Vector3f((float)sin(f_side_angle + f_pi * .5f), 0,
			(float)cos(f_side_angle + f_pi * .5f));
		p_vert_ptr->v_bitangent = Vector3f(0, -1, 0);
		// normal vectors

		if(!n_vertex_index && n_index % 12)
			n_vertex_index = 6;
		p_vert_ptr->v_texcoord = Vector2f(n_vertex_index, (float)(n_index >= 12) *
			.5f / (f_rod_radius / f_rod_height));
		// texcoords (.5f is side texture aspect, f_rod_radius / f_rod_height is rod side aspect)
	}}
	// sides

	// have array with pos3, normal3, tangent3, bitangent3, texcoord2

	float *p_interleaved_ptr = m_p_rod_vertex_array;
	for(TVertex *p_vert_ptr = p_vertex, *p_end = p_vertex + 30;
	   p_vert_ptr < p_end; ++ p_vert_ptr) {
		*p_interleaved_ptr ++ = p_vert_ptr->v_texcoord.x;
		*p_interleaved_ptr ++ = p_vert_ptr->v_texcoord.y;
		// T2F

		*p_interleaved_ptr ++ = -p_vert_ptr->v_tangent.x;
		*p_interleaved_ptr ++ = -p_vert_ptr->v_tangent.y;
		*p_interleaved_ptr ++ = -p_vert_ptr->v_tangent.z;
		*p_interleaved_ptr ++ = 0;
		// C4F

		*p_interleaved_ptr ++ = p_vert_ptr->v_normal.x;
		*p_interleaved_ptr ++ = p_vert_ptr->v_normal.y;
		*p_interleaved_ptr ++ = p_vert_ptr->v_normal.z;
		// N3F

		*p_interleaved_ptr ++ = p_vert_ptr->v_position.x;
		*p_interleaved_ptr ++ = p_vert_ptr->v_position.y;
		*p_interleaved_ptr ++ = p_vert_ptr->v_position.z;
		// V3F
	}
	// fill interleaved array with format: GL_T2F_C4F_N3F_V3F
}

void CHexLattice::Draw_Rod(CGLState *p_state, int n_x, int n_z, float f_y)
{
	glPushMatrix();

	glTranslatef(f_rod_radius * .9f * n_x, f_y * .5f,
		f_rod_radius * 1.1f * (.5f * (n_x & 1) + 3 * n_z) + f_rod_radius * (n_x & 1));
	// translate to proper position

	const GLuint n_dl = 1;

	if(!m_b_dl_ready) {
		glNewList(n_dl, GL_COMPILE_AND_EXECUTE);

		glInterleavedArrays(GL_T2F_C4F_N3F_V3F, 12 * sizeof(float), m_p_rod_vertex_array);

		p_state->ActiveTextureUnit(1);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("specular"));
		m_p_rod_top_specular->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(2);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("local"));
		m_p_rod_top_local->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(0);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("diffuse"));
		m_p_rod_top_diffuse->Bind_Enable(p_state);
		glDrawElements(GL_TRIANGLE_FAN, 6, GL_UNSIGNED_INT, m_p_rod_index_array_top);
		// draw top

		p_state->ActiveTextureUnit(1);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("specular"));
		m_p_rod_side_specular->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(2);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("local"));
		m_p_rod_side_local->Bind_Enable(p_state);
		p_state->ActiveTextureUnit(0);
		//p_state->ActiveTextureUnit(m_p_local_bump_shader->n_TextureUnit("diffuse"));
		m_p_rod_side_diffuse->Bind_Enable(p_state);
		glDrawElements(GL_QUADS, 24, GL_UNSIGNED_INT, m_p_rod_index_array_side);
		// draw sides

		glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		glDisableClientState(GL_COLOR_ARRAY);
		glDisableClientState(GL_NORMAL_ARRAY);
		glDisableClientState(GL_VERTEX_ARRAY);
		// disable vertex array state

		glEndList();
		m_b_dl_ready = true;
	} else {
		glCallList(n_dl);
		for(int i = 2; i >= 0; -- i) {
			p_state->ActiveTextureUnit(i);
			p_state->EnableTexture2D();
			p_state->DisableTexture2D();
			p_state->BindTexture2D(1);
			p_state->BindTexture2D(0); // texture state was dammaged by use of display list
		}
		//glColor3f(0, 0, 0);
		glColor3f(1, 1, 1); // color state was dammaged by vertex array
	}

	glPopMatrix();
}

/*
 *								=== ~CHexLattice ===
 */

/*
 *								=== CEffect2 ===
 */

/*
 *	virtual const char *CEffect2::p_s_Name()
 *		- returns short effect name (c-like identifier, name of effect class for example)
 * /
const char *CEffect2::p_s_Name()
{
	return "";
}

/*
 *	virtual const char *CEffect2::p_s_LongName()
 *		- returns full effect name
 * /
const char *CEffect2::p_s_LongName()
{
	return "";
}

/*
 *	virtual bool CEffect2::b_Supported()
 *		- returns true in case effect is supported by OpenGL, otherwise false
 * /
bool CEffect2::b_Supported()
{
	return true;
}

/*
 *	virtual bool CEffect2::Init(CGLState *p_state)
 *		- loads all effect resources, creates OpenGL buffers, compiles shaders, etc
 *		- returns true on success, false on failure
 * /
bool CEffect2::Init(CGLState *p_state)
{
	return true;
}

/*
 *	virtual bool CEffect2::b_Status()
 *		- returns true in case effect is loaded and ready for rendering, otherwise false
 * /
bool CEffect2::b_Status()
{
	return true;
}

/*
 *	virtual bool CEffect2::Shutdown(CGLState *p_state)
 *		- shutdown effect
 *		- returns true in case no errors occured (even in case effect was not initialized)
 *		  otherwise returns false
 * /
bool CEffect2::Shutdown(CGLState *p_state)
{
	return true;
}

/*
 *	virtual float CEffect2::f_Length()
 *		- returns effect length in seconds
 * /
float CEffect2::f_Length()
{
	return 50;
}

/*
 *	virtual bool CEffect2::Render(CGLState *p_state, float f_time)
 *		- renders a single frame into current framebuffer; f_time is frame time,
 *		  it can be out of <0, f_Length()> range, it's up to effect on how to handle that
 *		  (repeat itself, fade out, render nothing ...)
 * /
bool CEffect2::Render(CGLState *p_state, float f_time)
{
	return true;
}

/*
 *								=== ~CEffect2 ===
 */

