/*
								+---------------------------------+
								|                                 |
								|   *** simple win32 window ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Main.cpp             |
								|                                 |
								+---------------------------------+
*/

#include "../UberLame_src/NewFix.h"
#include "../UberLame_src/CallStack.h"
#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "../UberLame_src/OpenGL20.h"
#include "../UberLame_src/OpenGLDrv.h"
#include "../UberLame_src/OpenGLState.h"
#include "../UberLame_src/Timer.h"
#include "../UberLame_src/Mersene.h"
#include "../UberLame_src/Vector.h"
#include "../UberLame_src/Texture.h"
#include "../UberLame_src/Shader2.h"
#include "../UberLame_src/ShaderSys.h"
#include "../UberLame_src/MinMax.h"
#include "Effect.h"
#include <multimon.h>
#include "Monitors.h"

//#define __CATCH_ON_PAINT
// define this if you want to use WM_PAINT for drawing into the window
// (then simply put your drawing code inside OnPaint())

#define __USE_PEEK_MESSAGE
// define this if you want to use the OnIdle() function

static int n_width = 640;
static int n_height = 480;
// window size constants

static const bool b_single_monitor = true;
static int n_monitor_no = 0;

static const bool b_fullscreen = false;
static const bool b_centered = false;
static const bool b_topmost = false;
// window placement constants

static const bool b_window_frame = false;
static const bool b_show_cursor = false;
// window style constants

static const char *p_s_window_name = "HexLattice saver";
static const char *p_s_class_name = "HexLatSaver";
// window name constants

#ifdef __USE_PEEK_MESSAGE
static bool b_running = true;
#endif
static bool b_focus = true;
// window state variables

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param);
#ifdef __USE_PEEK_MESSAGE
void OnIdle(HWND h_wnd);
#endif
#ifdef __CATCH_ON_PAINT
void OnPaint(HWND h_wnd, HDC h_dc, int n_width, int n_height);
#endif
void Render(bool b_switch_context, int n_mon_index);
// function prototypes

static CGLState **p_state = 0;
static CGLDriver *p_ogl_drv;
static HWND *p_hwnd;
static CSmokeBackground *p_background_eff;
static CHexLattice *p_lattice_eff;
static CTimer timer; // animation timer

static bool b_shutdown = false;
static bool b_mouse_tracked = false;
static int n_mouse_x, n_mouse_y;

int WINAPI WinMain(HINSTANCE h_instance, HINSTANCE h_prev_instance, PSTR p_s_cmd_line, int n_cmd_show)
{
	if(FindWindow(p_s_class_name, 0) || !strstr(p_s_cmd_line, "/s") && !strstr(p_s_cmd_line, "/S"))
		return 0;
	// try to find running screensaver window (hPrevInstance doesn't work)

	TMultimonInfo t_monitor_info;
	if(!t_monitor_info.Update_Monitors())
		return -1;

	const TMonitorInfo *p_monitor_begin = t_monitor_info.p_Monitor(0);
	int n_monitor_num = t_monitor_info.n_Monitor_Num();
	if(b_single_monitor) {
		p_monitor_begin += min(n_monitor_no, n_monitor_num - 1);
		n_monitor_num = 1;
	}
	const TMonitorInfo *p_monitor_end = p_monitor_begin + n_monitor_num;
	// get monitor list

	if(!(p_hwnd = new(std::nothrow) HWND[n_monitor_num]))
		return -1;
	if(!(p_background_eff = new(std::nothrow) CSmokeBackground[n_monitor_num])) {
		delete[] p_hwnd;
		return -1;
	}
	if(!(p_lattice_eff = new(std::nothrow) CHexLattice[n_monitor_num])) {
		delete[] p_background_eff;
		delete[] p_hwnd;
		return -1;
	}
	if(!(p_state = new(std::nothrow) CGLState*[n_monitor_num])) {
		delete[] p_background_eff;
		delete[] p_lattice_eff;
		delete[] p_hwnd;
		return -1;
	}
	if(!(p_ogl_drv = new(std::nothrow) CGLDriver[n_monitor_num])) {
		delete[] p_background_eff;
		delete[] p_lattice_eff;
		delete[] p_state;
		delete[] p_hwnd;
		return -1;
	}

	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = WndProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = h_instance;
	t_wnd_class.hIcon = LoadIcon(h_instance, IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(h_instance, IDC_ARROW);
	t_wnd_class.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_class_name;
	t_wnd_class.hIconSm = LoadIcon(h_instance, IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return -1;

	int n_x, n_y;
	if(b_fullscreen) {
		n_x = 0;
		n_y = 0;
		n_width = GetSystemMetrics(SM_CXSCREEN);
		n_height = GetSystemMetrics(SM_CYSCREEN);
	} else if(b_centered) {
		if(n_width == CW_USEDEFAULT)
			n_width = 640;
		if(n_height == CW_USEDEFAULT)
			n_height = 480;
		n_x = (GetSystemMetrics(SM_CXSCREEN) - n_width) / 2;
		n_y = (GetSystemMetrics(SM_CYSCREEN) - n_height) / 2;
	} else {
		n_x = CW_USEDEFAULT;
		n_y = CW_USEDEFAULT;
	}

	RECT t_rect;
	t_rect.left = 0;
	t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	AdjustWindowRectEx(&t_rect, (b_window_frame)? WS_OVERLAPPEDWINDOW : WS_POPUP,
		FALSE, WS_EX_APPWINDOW | ((b_window_frame)? WS_EX_WINDOWEDGE : 0));

	for(const TMonitorInfo *p_monitor = p_monitor_begin; p_monitor != p_monitor_end; ++ p_monitor) {
		char p_s_wnd_name[192];
		sprintf(p_s_wnd_name, "HexSaver @ \'%s\'", p_monitor->p_s_name);

		int n_index = p_monitor - p_monitor_begin;
		if(!(p_hwnd[n_index] = CreateWindowEx((b_topmost)? WS_EX_TOPMOST : 0, p_s_class_name, p_s_window_name,
		   (b_window_frame)? WS_OVERLAPPEDWINDOW : WS_POPUP,
		   p_monitor->n_x, p_monitor->n_y, p_monitor->n_width, p_monitor->n_height,
		   //50, 50 + 400 * n_index, 400, 300,
		   NULL, NULL, h_instance, NULL))) {
			delete[] p_background_eff;
			delete[] p_lattice_eff;
			delete[] p_state;
			delete[] p_ogl_drv;
			delete[] p_hwnd;
			return -1;
		}
		ShowWindow(p_hwnd[n_index], n_cmd_show);
		UpdateWindow(p_hwnd[n_index]);

		if(!p_ogl_drv[n_index].Init(p_hwnd[n_index], p_monitor->n_width,
		   p_monitor->n_height, 24, 0, 0, false)) {
			delete[] p_background_eff;
			delete[] p_lattice_eff;
			delete[] p_state;
			delete[] p_ogl_drv;
			delete[] p_hwnd;
			return -1;
		}
		// init open-gl

		p_ogl_drv[n_index].MakeCurrent();

		if(!(p_state[n_index] = new(std::nothrow) CGLState(/*wglGetCurrentDC(), wglGetCurrentContext()*/))) {
			delete[] p_background_eff;
			delete[] p_lattice_eff;
			delete[] p_state;
			delete[] p_ogl_drv;
			delete[] p_hwnd;
			return -1;
		}
		// state man

		glEnable(GL_LINE_SMOOTH);
		glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

		p_state[n_index]->ShadeModel(GL_SMOOTH);
		p_state[n_index]->EnableDepthTest();

		if(!p_background_eff[n_index].Init(p_state[n_index], p_monitor->n_width, p_monitor->n_height) ||
		   !p_lattice_eff[n_index].Init(p_state[n_index], p_monitor->n_width, p_monitor->n_height)) {
			delete[] p_background_eff;
			delete[] p_lattice_eff;
			delete[] p_state;
			delete[] p_ogl_drv;
			delete[] p_hwnd;
			return -1;
		}
	}

	if(!b_show_cursor)
		ShowCursor(false);

	MSG t_msg;
	for(int n_window_num = n_monitor_num; n_window_num;) {
		if(b_shutdown) {
			b_shutdown = false;
			if(p_ogl_drv) {
				for(int i = 0; i < n_monitor_num; ++ i) {
					if(p_ogl_drv[i].MakeCurrent()) { // so we can delete textures
						p_background_eff[i].Shutdown(p_state[i]);
						p_lattice_eff[i].Shutdown(p_state[i]);
						delete p_state[i];
						p_ogl_drv[i].Shutdown();
					}
				}
				delete[] p_ogl_drv;
				p_ogl_drv = 0;
			}
			// terminate open-gl on all windows at once

			for(int i = 0; i < n_monitor_num; ++ i)
				PostMessage(p_hwnd[i], WM_CLOSE, 0, 0);
			// close windows
		}
		if(PeekMessage(&t_msg, NULL, 0, 0, PM_REMOVE)) { 
			if(t_msg.message == WM_CLOSE)
				n_window_num --;
            TranslateMessage(&t_msg);
            DispatchMessage(&t_msg);
        } else if(p_ogl_drv) {
			for(const TMonitorInfo *p_monitor = p_monitor_begin; p_monitor != p_monitor_end; ++ p_monitor) {
				if(!p_monitor->b_on || !p_monitor->b_valid)
					continue;
				// render only onto monitors switched on

				Render(n_monitor_num > 1, p_monitor - p_monitor_begin);
				// render (not using h_wnd anymore)
			}
		}
	}

	delete[] p_background_eff;
	delete[] p_lattice_eff;
	delete[] p_state;
	delete[] p_ogl_drv;
	delete[] p_hwnd;

	if(!b_show_cursor)
		ShowCursor(true);

	return 0;
}

LRESULT CALLBACK WndProc(HWND h_wnd, UINT n_msg, WPARAM n_w_param, LPARAM n_l_param)
{
	switch(n_msg) {
	case WM_SETFOCUS:
		b_focus = true;
		break;

	case WM_KILLFOCUS:
		b_focus = false;
		break;

	case WM_KEYDOWN:
		b_shutdown = true;//SendMessage(h_wnd, WM_CLOSE, 0, 0);
		break;

	case WM_MOUSEMOVE:
		{
			POINT t_point;
			GetCursorPos(&t_point);
			if(!b_mouse_tracked) {
				b_mouse_tracked = true;
				n_mouse_x = t_point.x;
				n_mouse_y = t_point.y;
			}
			if(abs(t_point.x - n_mouse_x) > 5 || abs(t_point.y - n_mouse_y) > 5)
				b_shutdown = true;//SendMessage(h_wnd, WM_CLOSE, 0, 0);
		}
		break;

#ifdef __CATCH_ON_PAINT
	case WM_PAINT:
		{
			PAINTSTRUCT t_ps;
			HDC h_dc;
			h_dc = BeginPaint(h_wnd, &t_ps);

			RECT t_rect;
			GetClientRect(h_wnd, &t_rect);

			OnPaint(h_wnd, h_dc, t_rect.right - t_rect.left, t_rect.bottom - t_rect.top);

			EndPaint(h_wnd, &t_ps);
		}
		return 0;
#endif //__CATCH_ON_PAINT

	case WM_DESTROY:
		PostQuitMessage(0);
#ifdef __USE_PEEK_MESSAGE
		b_running = false;
#endif
		return 0;
	}

	return DefWindowProc(h_wnd, n_msg, n_w_param, n_l_param) ;
}

#ifdef __USE_PEEK_MESSAGE

/*
 *	void OnIdle(HWND h_wnd)
 *		- called when window is not receiving any messages
 *		- note __USE_PEEK_MESSAGE must be defined
 */
void OnIdle(HWND h_wnd)
{
}

#endif //__USE_PEEK_MESSAGE

#ifdef __CATCH_ON_PAINT

/*
 *	void OnPaint(HWND h_wnd, HDC h_dc, int n_width, int n_height)
 *		- called when WM_PAINT message arrives
 *		- h_wnd is window handle, h_dc is device context, n_width and n_height
 *		  are dimensions of client area
 *		- note __CATCH_ON_PAINT must be defined
 */
void OnPaint(HWND h_wnd, HDC h_dc, int n_width, int n_height)
{
	// todo - fill in your paint code here
}

#endif //__CATCH_ON_PAINT

/*
 *								=== OpenGL ===
 */

/*
 *	void Render(bool b_switch_context, int n_mon_index)
 *		- render OpenGL frame
 */
void Render(bool b_switch_context, int n_mon_index)
{
	if(b_switch_context) {
		p_ogl_drv[n_mon_index].MakeCurrent();
		p_ogl_drv[n_mon_index].ResetViewport();
	}

	p_state[n_mon_index]->ClearColor4f(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if(!p_background_eff[n_mon_index].b_Status() || !p_lattice_eff[n_mon_index].b_Status())
		return;

	if(!p_background_eff[n_mon_index].Render(p_state[n_mon_index], timer.f_Time()) ||
	   !p_lattice_eff[n_mon_index].Render(p_state[n_mon_index], timer.f_Time()))
		b_shutdown = true;

	p_ogl_drv[n_mon_index].Blit();
}

/*
 *								=== ~OpenGL ===
 */

